from sail import cli, util

import click, pathlib, json
import os, re
import requests
import time
import click.utils

import curses
import textwrap
from urllib.parse import urlparse
from datetime import datetime
import subprocess
import shlex

# shortcuts:
# sail profile .profiles/some-file.json
# sail profile https://example.org/

# sail profile open .profiles/something.json
# sail profile run https://example.org
# sail profile curl -H 'Some: header' -XPOST https://example.org
# sail profile key --curl
# sail profile download /var/www/profiles/filename.xhprof

class ProfilerCmd(click.Group):
	def resolve_command(self, ctx, args):
		cmd_name = click.utils.make_str(args[0])

		if cmd_name.startswith('http://') or cmd_name.startswith('https://'):
			args.insert(0, 'run')

		path = pathlib.Path(cmd_name)
		if path.exists() and path.is_file():
			args.insert(0, 'open')

		return super().resolve_command(ctx, args)

@cli.group(cls=ProfilerCmd)
@click.pass_context
def profile(ctx):
	'''Run the profiler to find application performance bottlenecks'''
	pass

@profile.command()
@click.argument('path', nargs=1)
def open(path):
	'''Open the profile browser with the specified JSON file'''
	path = pathlib.Path(path)
	if not path.exists() or not path.is_file():
		raise click.ClickException('The profile file is invalid or does not exist')

	with path.open('r') as f:
		try:
			profile_data = json.load(f)
			xhprof_data = profile_data['xhprof']
		except:
			raise click.ClickException('This profile file is invalid, invalid or incomplete JSON')

	del profile_data['xhprof']

	totals = {'ct': 0, 'wt': 0, 'ut': 0, 'st': 0, 'cpu': 0, 'mu': 0, 'pmu': 0, 'samples': 0,
		'queries': 0, 'http_reqs': 0, 'timestamp': profile_data['timestamp'],
		'method': profile_data['method'], 'host': profile_data['host'],
		'request_uri': profile_data['request_uri']}

	metrics = []
	for key in totals.keys():
		if key in xhprof_data['main()']:
			metrics.append(key)

	data = {}

	# Compute inclusive times
	for func, info in xhprof_data.items():
		try:
			parent, child = func.split('==>')
		except:
			child = func
			parent = None

		if child not in data:
			data[child] = {'ct': 0, 'wt': 0, 'ut': 0, 'st': 0, 'cpu': 0,
				'mu': 0, 'pmu': 0, 'samples': 0, 'parents': [], 'children': []}

		for metric in metrics:
			data[child][metric] += info[metric]

		# More counts
		func = child.split('#', 1)[0]
		if func in ['mysqli_query', 'mysql_query', 'mysqli::query']:
			totals['queries'] += data[child]['ct']
		elif func in ['curl_exec']:
			totals['http_reqs'] += data[child]['ct']

	for metric in metrics:
		totals[metric] = data['main()'][metric]

	# Reset the calls, we'll count them again
	totals['ct'] = 0

	for child, info in data.items():
		totals['ct'] += info['ct']
		for metric in metrics:
			data[child]['excl_' + metric] = info[metric]

	for func, info in xhprof_data.items():
		try:
			parent, child = func.split('==>')
		except:
			child = func
			parent = None

		if not parent:
			continue

		data[child]['parents'].append(parent)
		data[parent]['children'].append(child)

		for metric in metrics:
			data[parent]['excl_' + metric] -= info[metric]

	os.environ.setdefault('ESCDELAY', '10')
	curses.wrapper(_browser, data=data, totals=totals)

@profile.command()
@click.argument('url', nargs=1)
@click.pass_context
def run(ctx, url):
	'''Run the profiler on a URL, download and open the results'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	if 'profile_key' not in sail_config:
		raise click.ClickException('Profile key not found in .sail/config.json')

	url = urlparse(url)
	host = '%s.sailed.io' % sail_config['app_id']
	query = url.query
	nocache = 'SAIL_NO_CACHE=%d' % time.time()
	query = nocache if not query else query + '&' + nocache
	query = '?' + query
	url_path = url.path if url.path else '/'

	click.echo('# Profiling')
	click.echo('- Server: %s' % host)
	click.echo('- Host: %s' % url.netloc)
	click.echo('- Request: GET %s%s' % (url_path, query))

	headers = {
		'Host': url.netloc,
		'X-Sail-Profile': sail_config['profile_key'],
	}

	request = requests.Request('GET', '%s://%s%s%s' % (url.scheme, host, url.path, query), headers=headers)
	request = request.prepare()
	session = requests.Session()

	try:
		response = session.send(request)
	except:
		raise click.ClickException('Could not make profiling request.')

	if 'X-Sail-Profile' not in response.headers:
		raise click.ClickException('X-Sail-Profile header not found in response. Check your profile key.')

	filename = response.headers['X-Sail-Profile']
	profile_path = ctx.invoke(download, path=filename)
	return ctx.invoke(open, path=profile_path)

@profile.command()
@click.option('--header', is_flag=True, help="Provide the result as an HTTP header string")
def key(header):
	'''Show the profiling secret key'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	if 'profile_key' not in sail_config:
		raise click.ClickException('Profile key not found in .sail/config.json')

	if header:
		click.echo('X-Sail-Profile: %s' % sail_config['profile_key'], nl=False)
		return

	click.echo(sail_config['profile_key'], nl=False)

@profile.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('command', nargs=-1)
@click.pass_context
def curl(ctx, command):
	'''Wrapper for the curl command which adds an X-Sail-Profile header'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	if 'profile_key' not in sail_config:
		raise click.ClickException('Profile key not found in .sail/config.json')

	command = list(command)
	command = ['-s', '-v', '-H', 'X-Sail-Profile: %s' % sail_config['profile_key']] + command
	click.echo('# Running cURL with profiling headers', err=True)

	# TODO: Maybe add the SAIL_NOCACHE query var
	p = subprocess.Popen(['curl'] + command, stdout=subprocess.PIPE,
		stderr=subprocess.PIPE, encoding='utf8')

	stdout, stderr = p.communicate()

	if p.returncode != 0:
		raise click.ClickException('An error occurred in cURL. Please try again.')

	matches = re.search(r'^<\s*X-Sail-Profile: (.+)$', stderr, re.MULTILINE)
	if not matches:
		raise click.ClickException('Could not find profile filename from headers')

	filename = matches.group(1)
	profile_path = ctx.invoke(download, path=filename)
	return ctx.invoke(open, path=profile_path)

@profile.command()
@click.argument('path', nargs=1)
def download(path):
	'''Download a profile JSON from the production server'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	if 'profile_key' not in sail_config:
		raise click.ClickException('Profile key not found in .sail/config.json')

	profiles_dir = pathlib.Path(root + '/.profiles')
	profiles_dir.mkdir(parents=True, exist_ok=True)

	dest_filename = datetime.now().strftime('%Y-%m-%d-%H%M%S.xhprof.json')
	click.echo('- Downloading profile from %s' % path)

	args = ['-t']
	source = 'root@%s.sailed.io:%s' % (sail_config['app_id'], path)
	destination = '%s/%s' % (profiles_dir, dest_filename)
	returncode, stdout, stderr = util.rsync(args, source, destination, default_filters=False)

	if returncode != 0:
		raise click.ClickException('An error occurred in rsync. Please try again.')

	click.echo('- Cleaning up production')

	p = subprocess.Popen(['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'rm %s' % path
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		raise click.ClickException('An error occurred in SSH. Please try again.')

	click.echo('- Profile saved to .profiles/%s' % dest_filename)
	return profiles_dir / dest_filename

def _render_summary(pad, totals):
	run_id = datetime.fromtimestamp(totals['timestamp']).strftime('%Y-%m-%d %H:%M:%S')

	pad.addstr(0, 0, 'Run: ', curses.color_pair(1))
	pad.addstr(run_id, curses.color_pair(2))
	pad.addstr(' Wall Time: ', curses.color_pair(1))
	pad.addstr('{:,} µs'.format(totals['wt']), curses.color_pair(2))
	pad.addstr(' Peak Memory: ', curses.color_pair(1))
	pad.insstr('{:,.2f} MiB'.format(totals['pmu']/1024/1024), curses.color_pair(2))

	request_uri = totals['request_uri']
	request_uri = re.sub(r'&?SAIL_NO_CACHE=[0-9]+', '', request_uri)
	request_uri = request_uri.rstrip('?')
	if request_uri == '/':
		request_uri = ''

	url = totals['host'] + request_uri
	pad.addstr(1, 0, 'URL: ', curses.color_pair(1))
	pad.addstr(1, 5, url, curses.color_pair(2))

	pad.addstr(2, 0, 'Method: ', curses.color_pair(1))
	pad.addstr(totals['method'], curses.color_pair(2))

	pad.addstr(' Function Calls: ', curses.color_pair(1))
	pad.addstr('{:,}'.format(totals['ct']), curses.color_pair(2))
	pad.addstr(' Queries: ', curses.color_pair(1))
	pad.addstr('{:,}'.format(totals['queries']), curses.color_pair(2))
	pad.addstr(' HTTP Reqs: ', curses.color_pair(1))
	pad.addstr('{:,}'.format(totals['http_reqs']), curses.color_pair(2))

def _render_listview(pad, columns, data, cols, selected = 0):
	y = -1
	for entry in data:
		y += 1

		if y == selected:
			pad.attron(curses.color_pair(4) if curses.has_colors() else curses.A_REVERSE)
		else:
			pad.attroff(curses.color_pair(4) if curses.has_colors() else curses.A_REVERSE)

		if 'header' in entry:
			pad.attron(curses.color_pair(3) if curses.has_colors() else curses.A_REVERSE)
			pad.hline(y, 0, ' ', cols)
			pad.addstr(y, 0, '  ' + entry['header'])
			pad.attroff(curses.color_pair(3) if curses.has_colors() else curses.A_REVERSE)
			continue

		if 'space' in entry:
			pad.hline(y, 0, ' ', cols)
			continue

		if 'meta' in entry:
			if y != selected:
				pad.attron(curses.color_pair(5))
			pad.hline(y, 0, ' ', cols)
			pad.insstr(y, 0, '  ' + entry['meta'])
			if y != selected:
				pad.attroff(curses.color_pair(5))

			continue

		label = entry['label']
		args = entry.get('args')

		pad.hline(y, 0, ' ', cols)
		pad.addstr(y, 0, '  ' + label)
		if args:
			if y != selected:
				pad.attron(curses.color_pair(5))
			pad.insstr(' => ' + args)
			if y != selected:
				pad.attroff(curses.color_pair(5))
		for column in reversed(columns):
			pad.insstr(y, cols - sum([i['width'] for i in columns]) - 2, '{:,}'.format(entry[column['key']]).rjust(column['width']))

		pad.insstr(y, cols-2, '  ')

def _render_footer(pad, selected, max, cols):
	max -= 1 # selected is 0-based
	pad.attron(curses.color_pair(1))
	pad.hline(0, 0, '-', cols)
	label = ' %d/%d %d%% ' % (selected, max, selected/max * 100)
	pad.insstr(0, cols - len(label) - 1, label )
	pad.attroff(curses.color_pair(1))

def _render_sticky_header(pad, columns, data, cols, offset_y):
	header = None
	i = offset_y
	while not header and i >= 0:
		if 'header' in data[i]:
			header = data[i]

		i -= 1

	if not header:
		return

	pad.attron(curses.color_pair(3) if curses.has_colors() else curses.A_REVERSE)
	pad.hline(0, 0, ' ', cols)
	# pad.addstr(0, 0, 'Off: %d, sel: %d, i: %d' % (offset_y, selected, i))
	pad.addstr(0, 0, '  ' + header['header'])

	for column in reversed(columns):
		label = column['label']
		if 'sort' in column:
			label = '>' + label
		else:
			label = ' ' + label

		pad.insstr(0, cols - sum([i['width'] for i in columns]) - 2, label.rjust(column['width']))

	pad.attroff(curses.color_pair(3) if curses.has_colors() else curses.A_REVERSE)

def _render_view_symbol(stdscr, data, totals, symbol, selected=1, sort=2):
	rows, cols = stdscr.getmaxyx()

	columns = _columns();
	for column in columns:
		column['width'] = max(len(column['label']), len('{:,}'.format(max([i[column['key']] for _, i in data.items()])))) + 2

	sort_key = columns[sort]['key']
	columns[sort]['sort'] = True
	listview_data = [{'header': 'Function'}]

	entry = data[symbol]

	item = {'function': symbol, 'label': symbol}

	meta = None
	expand_meta_for = [
		'mysqli_query',
		'mysql_query',
		'curl_exec',
		'mysqli::query',
		'{closure}',
	]

	if '#' in symbol:
		item['label'], args = symbol.split('#', 1)

		if item['label'] in expand_meta_for:
			meta = []
			for line in textwrap.wrap(args, cols - 4):
				meta.append({'meta': line})
		else:
			item['args'] = args

	for col in columns:
		item[col['key']] = entry[col['key']]

	listview_data.append(item)
	if meta:
		listview_data.append({'space': ''})
		listview_data.extend(meta)

	if len(set(entry['parents'])):
		listview_data.append({'space': ''})
		listview_data.append({'header': 'Parent functions'})
		parents = []

		for parent in entry['parents']:
			item = {'function': parent, 'label': parent}

			if '#' in parent:
				item['label'], item['args'] = parent.split('#', 1)

			for col in columns:
				item[col['key']] = data[parent][col['key']]

			parents.append(item)

		listview_data.extend(sorted(parents, key=lambda x: x[sort_key], reverse=True))

	if len(set(entry['children'])):
		listview_data.append({'space': ''})
		listview_data.append({'header': 'Child functions'})
		children = []

		for child in entry['children']:
			item = {'function': child, 'label': child}

			if '#' in child:
				item['label'], item['args'] = child.split('#', 1)

			for col in columns:
				item[col['key']] = data[child][col['key']]

			children.append(item)

		listview_data.extend(sorted(children, key=lambda x: x[sort_key], reverse=True))

	summary = curses.newpad(3, cols)
	listview = curses.newpad(len(listview_data), cols)
	sticky_header = curses.newpad(1, cols)
	footer = curses.newpad(1, cols)

	_render_summary(summary, totals)
	summary.refresh(0,0, 0,0, 4, cols - 1)

	visible = rows - 6
	offset_y = 0
	refresh = True

	while True:
		if selected > visible + offset_y:
			offset_y = selected - visible
		elif selected <= offset_y:
			offset_y = selected - 1

		if refresh:
			_render_listview(listview, columns, listview_data, cols, selected)
			listview.refresh(offset_y,0, 4,0, rows - 2, cols - 1)

			_render_sticky_header(sticky_header, columns, listview_data, cols, offset_y)
			sticky_header.refresh(0,0, 4,0, 6, cols -1)

			_render_footer(footer, selected, len(listview_data), cols)
			footer.refresh(0,0, rows-1,0, rows-1,cols-1)

		c = stdscr.getch()
		selected, refresh = _handle_scroll(c, listview_data, visible, selected, refresh)

		# Sorting
		if c == curses.KEY_RIGHT or c == ord('>'):
			next = min(sort + 1, len(columns) - 1)
			if next == sort:
				continue

			return ('sort', next)

		elif c == curses.KEY_LEFT or c == ord('<'):
			prev = max(sort - 1, 0)
			if prev == sort:
				continue

			return ('sort', prev)

		elif c == 27:
			return 'view_main'

		elif c == curses.KEY_BACKSPACE or c == 127 or c == '\b':
			return 'pop' # previous symbol or main

		elif c == curses.KEY_ENTER or c == 13 or c == 10:
			if 'function' not in listview_data[selected]:
				continue

			# First item is the function itself, don't select it.
			if selected == 1:
				continue

			return ('view_symbol', listview_data[selected]['function'], selected)

		elif c == ord('q'):
			return 'exit'

		elif c == curses.KEY_RESIZE:
			return 'resize'

def _render_view_main(stdscr, data, totals, selected=1, sort=2):
	rows, cols = stdscr.getmaxyx()

	columns = _columns()
	for column in columns:
		column['width'] = max(len(column['label']), len('{:,}'.format(max([i[column['key']] for _, i in data.items()])))) + 2

	listview_data = [
		{'header': 'Functions'},
	]

	sort_key = columns[sort]['key']
	columns[sort]['sort'] = True

	for func, info in sorted(data.items(), key=lambda x: x[1][sort_key], reverse=True):
		item = {'function': func, 'label': func}

		if '#' in func:
			item['label'], item['args'] = func.split('#', 1)

		for col in columns:
			item[col['key']] = info[col['key']]

		listview_data.append(item)

	summary = curses.newpad(3, cols)
	listview = curses.newpad(len(listview_data), cols)
	sticky_header = curses.newpad(1, cols)
	footer = curses.newpad(1, cols)

	_render_summary(summary, totals)
	summary.refresh(0,0, 0,0, 4, cols - 1)

	offset_y = 0
	visible = rows - 6
	refresh = True

	while True:
		if selected > visible + offset_y:
			offset_y = selected - visible
		elif selected <= offset_y:
			offset_y = selected - 1

		if refresh:
			_render_listview(listview, columns, listview_data, cols, selected)
			listview.refresh(offset_y,0, 4,0, rows-2,cols-1)

			_render_sticky_header(sticky_header, columns, listview_data, cols, offset_y)
			sticky_header.refresh(0,0, 4,0, 6,cols-1)

			_render_footer(footer, selected, len(listview_data), cols)
			footer.refresh(0,0, rows-1,0, rows-1,cols-1)
			refresh = False

		c = stdscr.getch()
		selected, refresh = _handle_scroll(c, listview_data, visible, selected, refresh)

		# Sorting
		if c == curses.KEY_RIGHT or c == ord('>'):
			next = min(sort + 1, len(columns) - 1)
			if next == sort:
				continue

			return ('sort', next)

		elif c == curses.KEY_LEFT or c == ord('<'):
			prev = max(sort - 1, 0)
			if prev == sort:
				continue

			return ('sort', prev)

		elif c == curses.KEY_ENTER or c == 13 or c == 10:
			if 'header' in listview_data[selected] or 'space' in listview_data[selected]:
				continue

			return ('view_symbol', listview_data[selected]['function'], selected)

		elif c == ord('q'):
			return 'exit'

		elif c == curses.KEY_RESIZE:
			return 'resize'

def _browser(stdscr, data, totals):
	try:
		curses.curs_set(False)
	except curses.error:
		pass

	current_view = _render_view_main
	args = [stdscr, data, totals]
	kwargs = {'selected': 1}
	sort = 2
	view_stack = []

	if curses.has_colors():
		curses.init_pair(1, 247, curses.COLOR_BLACK)
		curses.init_pair(2, 255, curses.COLOR_BLACK)
		curses.init_pair(3, 0, 245)
		curses.init_pair(4, 0, 255)
		curses.init_pair(5, 247, curses.COLOR_BLACK)

	while True:
		stdscr.refresh()
		kwargs['sort'] = sort

		try:
			r = current_view(*args, **kwargs)
		except curses.error:
			continue

		# Exit
		if r == 'exit' or r is None:
			print('\n')
			stdscr.erase()
			break

		# Escape back to main view
		if r == 'view_main':
			stdscr.erase()

			if view_stack:
				current_view, args, kwargs = view_stack.pop(0)
				view_stack = []
				continue

			current_view = _render_view_main
			args = [stdscr, data, totals]
			kwargs = {}
			continue

		if r == 'pop':
			stdscr.erase()

			# Pop back to the main view
			if len(view_stack) < 1:
				current_view = _render_view_main
				args = [stdscr, data, totals]
				continue

			# Pop back to the previous symbol
			current_view, args, kwargs = view_stack.pop()
			continue

		if r == 'resize':
			stdscr.erase()
			continue

		if r[0] == 'sort':
			_, sort = r
			kwargs['selected'] = 1
			continue

		# Enter symbol view
		if r[0] == 'view_symbol':
			view, symbol, selected = r

			# Push current settings to the stack
			kwargs['selected'] = selected
			view_stack.append((current_view, args, kwargs))

			current_view = _render_view_symbol
			args = [stdscr, data, totals]
			kwargs = {'symbol': symbol}
			stdscr.erase()
			continue

		stdscr.clear()
		stdscr.addstr(0, 0, repr(r))
		stdscr.refresh()
		stdscr.getch()
		break

def _columns():
	return [
		{'key': 'ct', 'label': 'Count', 'width': 10},
		{'key': 'wt', 'label': 'iWT', 'width': 10},
		{'key': 'excl_wt', 'label': 'eWT', 'width': 10},
		{'key': 'mu', 'label': 'iMEM', 'width': 10},
		{'key': 'excl_mu', 'label': 'eMEM', 'width': 10},
	]

def _is_valid(items, i):
	return 'header' not in items[i] and 'space' not in items[i]

def _closest_valid(items, current, step=1):
	i = current
	while True:
		i += step

		if i > len(items) - 1 or i < 0:
			return current

		if _is_valid(items, i):
			return i

def _handle_scroll(c, listview_data, visible, selected, refresh):
	_refresh = refresh

	if c == curses.KEY_UP:
		prev = _closest_valid(listview_data, selected, -1)
		_refresh = prev != selected
		selected = prev
	elif c == curses.KEY_DOWN:
		next = _closest_valid(listview_data, selected, +1)
		_refresh = next != selected
		selected = next
	elif c == curses.KEY_PPAGE:
		prev = max([selected - visible, 0])
		if not _is_valid(listview_data, prev):
			prev = _closest_valid(listview_data, prev, +1)
		_refresh = prev != selected
		selected = prev
	elif c == curses.KEY_NPAGE:
		next = min([selected + visible, len(listview_data) - 1])
		if not _is_valid(listview_data, next):
			next = _closest_valid(listview_data, next, -1)
		_refresh = next != selected
		selected = next
	elif c == curses.KEY_HOME:
		home = 0
		if not _is_valid(listview_data, home):
			home = _closest_valid(listview_data, home, +1)
		_refresh = home != selected
		selected = home
	elif c == curses.KEY_END:
		end = len(listview_data) - 1
		if not _is_valid(listview_data, end):
			end = _closest_valid(listview_data, end, -1)
		_refresh = end != selected
		selected = end

	return selected, refresh or _refresh



import sail

import time, pathlib, os
import json, click, requests, shlex, subprocess

_debug = False
def debug(set=None):
	'''Get or set debug mode'''
	global _debug

	if set is None:
		return _debug

	_debug = set
	return _debug

def dlog(line):
	'''Print debug output'''
	if not debug():
		return

	click.secho('(debug) %s' % line, fg='yellow')

def request(endpoint, **kwargs):
	'''Request the Sail API endpoint'''

	method = kwargs.get('method')
	params = kwargs.get('params', {})
	json = kwargs.get('json', {})
	anon = kwargs.get('anon', False)

	api_base = get_sail_default('api-base')
	if not api_base:
		api_base = sail.API_BASE

	headers = {}

	if not anon:
		sail_config = get_sail_config()
		headers = {
			'X-App-Id': sail_config['app_id'],
			'X-App-Secret': sail_config['secret'],
		}

	if not method:
		method = 'POST' if json else 'GET'

	url = '%s/%s' % (api_base.rstrip('/'), endpoint.lstrip('/'))

	if debug():
		safe_headers = headers.copy()
		if 'X-App-Secret' in safe_headers:
			safe_headers['X-App-Secret'] = '*'*8

		dlog('Requesting: [%s] %s' % (method, url))
		dlog('Request headers: %s' % repr(safe_headers))
		dlog('Request params: %s' % repr(params))
		dlog('Request json: %s' % repr(json))

	request = requests.Request(method, url, headers=headers, params=params)

	if method == 'POST' or json:
		request.json = json

	request = request.prepare()
	session = requests.Session()

	try:
		response = session.send(request)
	except Exception as e:
		dlog('Exception: %s' % repr(e))
		raise click.ClickException('Could not contact the Sail API. Try again later.')

	dlog('Response: [%d] %s' % (response.status_code, repr(response.text)))

	try:
		data = response.json()
	except Exception as e:
		dlog('Exception: %s' % repr(e))
		raise click.ClickException('Invalid response from the Sail API. Try again later.')

	if not response.ok:
		raise click.ClickException('API error: %s' % data.get('error'))

	return data

def wait_for_task(task_id, timeout=30, interval=1):
	feedback_processed = []
	data = {}

	def _wait():
		nonlocal data

		data = request('/status/%s/' % task_id)
		feedback = data.get('feedback', [])
		if feedback:
			for line in feedback:
				if line in feedback_processed:
					continue

				feedback_processed.append(line)
				click.echo('- %s' % line)

		if data.get('task_state') == 'failure':
			raise click.ClickException('Task state failure')

		return data.get('task_state') == 'success'

	wait(_wait, timeout=timeout, interval=interval)
	return data

def wait(condition, timeout=30, interval=1, *args):
	'''Wait for condition every interval or timeout.'''
	start = time.time()
	while not condition(*args):
		if time.time() - start > timeout:
			raise Exception('Timeout in util.wait()')

		time.sleep(interval)

def find_root():
	p = pathlib.Path()
	p = p.resolve()

	while p.parent and str(p) != '/':
		for x in p.iterdir():
			if x.stem == '.sail':
				return str(p)

		p = p.parent

def get_sail_config():
	root = find_root()

	if not root:
		raise click.ClickException('Could not parse .sail/config.json. If this is a new project run: sail init')

	with open(root + '/.sail/config.json') as f:
		config = json.load(f)

	if not config:
		raise click.ClickException('Could not parse .sail/config.json. If this is a new project run: sail init')

	return config

loader_i = 0

def loader():
	global loader_i

	frames = [
		"   ",
		".  ",
		".. ",
		"...",
	]

	click.echo(frames[loader_i % len(frames)] + '\r', nl=False)
	loader_i += 1
	time.sleep(.2)

def get_sail_default(name):
	'''Get a sail default config variable'''
	import pathlib, json
	filename = (pathlib.Path.home() / '.sail-defaults.json').resolve()
	data = {}

	try:
		with open(filename) as f:
			data = json.load(f)
			return data[name]
	except:
		return None

def rsync(args, source, destination, default_filters=True, extend_filters=[]):
	root = find_root()
	args = args[:]

	if debug():
		args.extend(['-v'])

	filters = []

	if type(default_filters) != bool:
		raise Exception('default_filters expected to be bool')

	if default_filters:
		filters = [
			'- /wp-content/debug.log',
			'- /wp-content/uploads',
			'- /wp-content/cache',
			'- /wp-content/upgrade',
		]

	if extend_filters:
		filters.extend(extend_filters)

	# Force exclude all dot-files
	filters.insert(0, '- .*')

	args.insert(0, 'rsync')
	ssh_args = ['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
	]

	args.extend(['-e', shlex.join(ssh_args)])

	# Add all filters in order
	for filter in filters:
		args.extend(['--filter', filter])

	args.extend([source, destination])

	if debug():
		dlog('Rsync: %s' % repr(args))

	# Download files FROM production
	p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf8')

	while True:
		try:
			stdout, stderr = p.communicate(timeout=0.2)
			break
		except subprocess.TimeoutExpired:
			loader()

	if debug():
		dlog('Rsync stdout: %s' % stdout)
		dlog('Rsync stderr: %s' % stderr)

	return (p.returncode, stdout, stderr)



from sail import cli, util

import click, pathlib, json
import os, shlex, subprocess
import webbrowser

@cli.command('config')
@click.argument('name', required=True, nargs=1)
@click.argument('value', required=False)
@click.option('--delete', is_flag=True)
def config(name, value=None, delete=False):
	'''Set reusable config variables'''
	valid_names = [
		'provider-token',
		'email',
		'api-base',
	]

	if name not in valid_names:
		raise click.ClickException('Invalid config name: %s' % name)

	filename = (pathlib.Path.home() / '.sail-defaults.json').resolve()
	data = {}

	try:
		with open(filename) as f:
			data = json.load(f)
	except:
		pass

	if not value and not delete:
		if name not in data:
			raise click.ClickException('The option is not set')

		click.echo(data[name])
		return

	if value and delete:
		raise click.ClickException('The --delete flag does not expect an option value')

	if delete and name not in data:
		raise click.ClickException('The option is not set, nothing to delete')

	if delete:
		del data[name]
		click.echo('Option %s deleted' % name)
	else:
		data[name] = value
		click.echo('Option %s set' % name)

	with open(filename, 'w+') as f:
		json.dump(data, f, indent='\t')

@cli.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('command', nargs=-1)
def wp(command):
	'''Run a WP-CLI command on the production host'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	command = shlex.join(command)

	click.echo('Spawning SSH and running WP-CLI on %s.sailed.io' % sail_config['app_id'], err=True)

	os.execlp('ssh', 'ssh', '-tt',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'docker exec -it sail sudo -u www-data bash -c "cd ~/public; wp %s"' % command
	)

@cli.command()
def admin():
	'''Open your default web browser to the wp-login.php location of your site'''
	root = util.find_root()
	sail_config = util.get_sail_config()
	webbrowser.open(sail_config['login_url'])

@cli.command()
@click.option('--nginx', is_flag=True)
@click.option('--php', is_flag=True)
@click.option('--nginx-access', is_flag=True)
@click.option('--nginx-error', '--nginx-errors', is_flag=True)
@click.option('--php-error', '--php-errors', is_flag=True)
@click.option('--postfix', '--mail', is_flag=True)
@click.option('--follow', '-f', is_flag=True)
@click.option('--lines', '-n', type=int)
def logs(nginx, php, nginx_access, nginx_error, php_error, postfix, follow, lines):
	'''Query and follow logs from the production server'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	click.echo('Querying logs on %s.sailed.io' % sail_config['app_id'])

	settings = []
	if nginx_access or nginx:
		settings.append('-t nginx_access')
	if nginx_error or nginx:
		settings.append('-t nginx_error')
	if php_error or php:
		settings.append('-t php')

	if follow:
		settings.append('--follow')
	elif lines:
		settings.append('--lines %d' % lines)
	else:
		settings.append('--lines 30')

	if postfix:
		settings.append('-t postfix/qmgr')
		settings.append('-t postfix/pickup')
		settings.append('-t postfix/master')
		settings.append('-t postfix/smtp')
		settings.append('-t postfix/cleanup')
		settings.append('-t postfix/postsuper')
		settings.append('-t postfix/postqueue')

	settings = ' '.join(settings)

	os.execlp('ssh', 'ssh', '-tt',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'journalctl --no-hostname --directory=/var/log/journal %s' % settings
	)

@cli.command()
@click.option('--root', '--host', is_flag=True, help='Login to the host (not the container) as the root user')
def ssh(root):
	'''Open an interactive SSH shell to the production container or host'''
	as_root = root
	root = util.find_root()
	sail_config = util.get_sail_config()

	# Container as www-data, or host as root
	command = ''
	if not as_root:
		command = 'docker exec -it sail sudo -u www-data bash -c "cd ~/public; bash"'

	click.echo('Spawning an interactive SSH shell for %s.sailed.io' % sail_config['app_id'])

	os.execlp('ssh', 'ssh', '-tt',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		command
	)



# Relative import
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).parent.parent.parent.resolve()))
from sail import cli

import io, os
import click
import unittest
import subprocess
import requests
import json
from click.testing import CliRunner
from unittest.mock import Mock, patch

# Skip some tests if true
work_in_progress = False

# Some commands use os.execlp to pass control to the
# child process, which stops test execution after the
# spawned process exits. We can mock it and just pipe
# to a regular subprocess.
def _execlp_side_effect(*args, **kwargs):
	print(subprocess.check_output(args[1:], encoding='utf8'))

_execlp = Mock(side_effect=_execlp_side_effect)

@patch('os.execlp', _execlp)
@unittest.skipIf(__name__ != '__main__', 'Slow test, must run explicitly')
class TestEnd2End(unittest.TestCase):
	@classmethod
	def setUpClass(cls):
		cls.runner = CliRunner(mix_stderr=False)
		cls.fs = cls.runner.isolated_filesystem()
		cls.fs.__enter__()
		cls.home = None

	@classmethod
	def tearDownClass(cls):
		cls.fs.__exit__(None, None, None)

	def setUp(self):
		self.home = self.__class__.home

	def test_000_config(self):
		api_base = 'http://127.0.0.1:5000/api/1.0/'

		# Allows running this on the production API server if explicitly asked.
		if 'SAIL_API_BASE' in os.environ:
			api_base = os.environ['SAIL_API_BASE']

		result = self.runner.invoke(cli, ['config', 'api-base', api_base])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Option api-base set', result.output)

	def test_001_init(self):
		result = self.runner.invoke(cli, ['init'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Success. The ship has sailed!', result.output)

	def test_002_wp_home(self):
		result = self.runner.invoke(cli, ['wp', 'option', 'get', 'home'])
		self.assertEqual(result.exit_code, 0)
		self.__class__.home = result.output.strip()

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_003_wp_media_import(self):
		result = self.runner.invoke(cli, ['wp', 'media', 'import', 'https://s.w.org/style/images/wp-header-logo.png'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Imported file', result.output)

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_004_download(self):
		result = self.runner.invoke(cli, ['download', '-y'])
		self.assertEqual(result.exit_code, 0)

		# Uploaded file shouldn't be here as we skip uploads by default
		found = subprocess.check_output(['find', '.', '-type', 'f', '-name', 'wp-header-logo.png'], encoding='utf8')
		self.assertNotIn('wp-header-logo.png', found)

		# With uploads now
		result = self.runner.invoke(cli, ['download', '-y', '--with-uploads'])
		self.assertEqual(result.exit_code, 0)

		# File should now be downloaded.
		found = subprocess.check_output(['find', '.', '-type', 'f', '-name', 'wp-header-logo.png'], encoding='utf8')
		self.assertIn('wp-header-logo.png', found)

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_005_php(self):
		with open('test.php', 'w') as f:
			f.write('<?php\necho "%s";' % self.home)

		# Deploy the test file
		result = self.runner.invoke(cli, ['deploy'])
		self.assertEqual(result.exit_code, 0)

		response = requests.get('%s/test.php' % self.home)
		self.assertTrue(response.ok)
		self.assertEqual(self.home, response.text)

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_006_deploy(self):
		with open('test1.php', 'w') as f:
			f.write('1')
		with open('test2.php', 'w') as f:
			f.write('2')

		# Make sure dry-run is working
		result = self.runner.invoke(cli, ['deploy', '--dry-run'])
		self.assertEqual(result.exit_code, 0)
		self.assertFalse(requests.get('%s/test1.php' % self.home).ok)

		# Test partial deploys
		result = self.runner.invoke(cli, ['deploy', 'test2.php'])
		self.assertEqual(result.exit_code, 0)
		self.assertFalse(requests.get('%s/test1.php' % self.home).ok)
		self.assertTrue(requests.get('%s/test2.php' % self.home).ok)

		result = self.runner.invoke(cli, ['deploy', 'test1.php'])
		self.assertEqual(result.exit_code, 0)
		self.assertTrue(requests.get('%s/test1.php' % self.home).ok)

		os.unlink('test1.php')
		result = self.runner.invoke(cli, ['deploy'])
		self.assertEqual(result.exit_code, 0)
		self.assertFalse(requests.get('%s/test1.php' % self.home).ok)
		self.assertTrue(requests.get('%s/test2.php' % self.home).ok)

		with open('test3.php', 'w') as f:
			f.write('3')
		with open('wp-content/test3.php', 'w') as f:
			f.write('also 3')
		with open('wp-content/plugins/test3.php', 'w') as f:
			f.write('three here as well')

		result = self.runner.invoke(cli, ['deploy', 'wp-content'])
		self.assertEqual(result.exit_code, 0)
		self.assertFalse(requests.get('%s/test3.php' % self.home).ok)
		self.assertTrue(requests.get('%s/wp-content/test3.php' % self.home).ok)
		self.assertTrue(requests.get('%s/wp-content/plugins/test3.php' % self.home).ok)

		os.unlink('wp-content/plugins/test3.php')
		result = self.runner.invoke(cli, ['deploy', 'wp-content/plugins'])
		self.assertEqual(result.exit_code, 0)
		self.assertFalse(requests.get('%s/test3.php' % self.home).ok)
		self.assertTrue(requests.get('%s/wp-content/test3.php' % self.home).ok)
		self.assertFalse(requests.get('%s/wp-content/plugins/test3.php' % self.home).ok)

		# Test wp-content/upload deploys --with-uploads
		with open('wp-content/uploads/test4.txt', 'w') as f:
			f.write('4')
		with open('wp-content/uploads/test5.txt', 'w') as f:
			f.write('5')

		result = self.runner.invoke(cli, ['deploy'])
		self.assertEqual(result.exit_code, 0)
		self.assertFalse(requests.get('%s/wp-content/uploads/test4.txt' % self.home).ok)
		self.assertFalse(requests.get('%s/wp-content/uploads/test5.txt' % self.home).ok)

		result = self.runner.invoke(cli, ['deploy', 'wp-content/uploads'])
		self.assertEqual(result.exit_code, 0)
		self.assertFalse(requests.get('%s/wp-content/uploads/test4.txt' % self.home).ok)
		self.assertFalse(requests.get('%s/wp-content/uploads/test5.txt' % self.home).ok)

		result = self.runner.invoke(cli, ['deploy', 'wp-content/uploads/test4.txt'])
		self.assertEqual(result.exit_code, 0)
		self.assertFalse(requests.get('%s/wp-content/uploads/test4.txt' % self.home).ok)
		self.assertFalse(requests.get('%s/wp-content/uploads/test5.txt' % self.home).ok)

		result = self.runner.invoke(cli, ['deploy', 'wp-content/uploads/test4.txt', '--with-uploads'])
		self.assertEqual(result.exit_code, 0)
		self.assertTrue(requests.get('%s/wp-content/uploads/test4.txt' % self.home).ok)
		self.assertFalse(requests.get('%s/wp-content/uploads/test5.txt' % self.home).ok)

		result = self.runner.invoke(cli, ['deploy', 'wp-content/uploads', '--with-uploads'])
		self.assertEqual(result.exit_code, 0)
		self.assertTrue(requests.get('%s/wp-content/uploads/test4.txt' % self.home).ok)
		self.assertTrue(requests.get('%s/wp-content/uploads/test5.txt' % self.home).ok)

		# Test both app and uploads deploys
		with open('test6.txt', 'w') as f:
			f.write('6')
		with open('wp-content/uploads/test7.txt', 'w') as f:
			f.write('7')
		with open('test8.txt', 'w') as f:
			f.write('8')
		with open('wp-content/uploads/test9.txt', 'w') as f:
			f.write('9')

		result = self.runner.invoke(cli, ['deploy', '--with-uploads', 'test6.txt', 'wp-content/uploads/test7.txt'])
		self.assertEqual(result.exit_code, 0)
		self.assertTrue(requests.get('%s/test6.txt' % self.home).ok)
		self.assertTrue(requests.get('%s/wp-content/uploads/test7.txt' % self.home).ok)
		self.assertFalse(requests.get('%s/test8.txt' % self.home).ok)
		self.assertFalse(requests.get('%s/wp-content/uploads/test9.txt' % self.home).ok)

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_007_download(self):
		files = [
			'download-1.txt',
			'download-2.txt',
			'wp-content/download-3.txt',
			'wp-content/plugins/download-4.txt',
			'wp-content/themes/download-5.txt',
			'wp-content/uploads/download-6.txt',
		]

		for filename in files:
			with open(filename, 'w') as f:
				f.write(filename)

		result = self.runner.invoke(cli, ['deploy', '--with-uploads'])
		self.assertEqual(result.exit_code, 0)

		# Delete all local files
		for filename in files:
			os.unlink(filename)

		should_exist = []
		def _assert_exists(files, should_exist):
			for filename in files:
				if filename not in should_exist:
					self.assertFalse(os.path.exists(filename))
				else:
					self.assertTrue(os.path.exists(filename))

		result = self.runner.invoke(cli, ['download', '--dry-run'])
		self.assertEqual(result.exit_code, 0)
		_assert_exists(files, should_exist)

		result = self.runner.invoke(cli, ['download', '-y', 'wp-content/plugins'])
		self.assertEqual(result.exit_code, 0)
		should_exist.append('wp-content/plugins/download-4.txt')
		_assert_exists(files, should_exist)

		result = self.runner.invoke(cli, ['download', '-y', 'wp-content'])
		self.assertEqual(result.exit_code, 0)
		should_exist.append('wp-content/themes/download-5.txt')
		should_exist.append('wp-content/download-3.txt')
		_assert_exists(files, should_exist)

		result = self.runner.invoke(cli, ['download', '-y', 'wp-content', '--with-uploads'])
		self.assertEqual(result.exit_code, 0)
		should_exist.append('wp-content/uploads/download-6.txt')
		_assert_exists(files, should_exist)

		# Test --delete
		with open('download-7.txt', 'w') as f:
			f.write('7')
		with open('wp-content/uploads/download-8.txt', 'w') as f:
			f.write('8')

		# Without the --delete flag, 7 and 8 should remain intact
		result = self.runner.invoke(cli, ['download', '-y', '--with-uploads'])
		self.assertEqual(result.exit_code, 0)
		should_exist.append('download-1.txt')
		should_exist.append('download-2.txt')
		_assert_exists(files, should_exist)

		# These still exists
		self.assertTrue(os.path.exists('download-7.txt'))
		self.assertTrue(os.path.exists('wp-content/uploads/download-8.txt'))

		# Add the --delete now, should be gone.
		result = self.runner.invoke(cli, ['download', '-y', '--with-uploads', '--delete'])
		self.assertEqual(result.exit_code, 0)
		self.assertFalse(os.path.exists('download-7.txt'))
		self.assertFalse(os.path.exists('wp-content/uploads/download-8.txt'))

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_008_blueprint_vars(self):
		result = self.runner.invoke(cli, ['blueprint', 'test_vars.yaml'], input='Simple\n\n\n\n')
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Test Simple', result.output)
		self.assertIn('Test Default [123]', result.output)
		self.assertIn('Test Type Boolean [True]', result.output)
		self.assertIn('Test Type Integer [123]', result.output)

		# Test command-line options
		result = self.runner.invoke(cli, ['blueprint', 'test_vars.yaml', '--test-simple=1',
			'--test-default=1', '--test-type-bool=yes', '--test-type-int=456'])

		self.assertEqual(result.exit_code, 0)
		self.assertNotIn('Test Simple', result.output)
		self.assertNotIn('Test Default', result.output)
		self.assertNotIn('Test Type Boolean', result.output)
		self.assertNotIn('Test Type Integer', result.output)

		result = self.runner.invoke(cli, ['wp', 'option', 'get', 'sail_vars_test', '--format=json'])
		self.assertEqual(result.exit_code, 0)
		vars = json.loads(result.output)

		self.assertEqual(vars['test_simple'], '1')
		self.assertEqual(vars['test_default'], '1')
		self.assertEqual(vars['test_type_bool'], True)
		self.assertEqual(vars['test_type_int'], 456)

		result = self.runner.invoke(cli, ['blueprint', 'test_vars.yaml', '--test-simple=1',
			'--test-default=1', '--test-type-bool=nah', '--test-type-int=789'])

		self.assertEqual(result.exit_code, 0)
		self.assertNotIn('Test Simple', result.output)
		self.assertNotIn('Test Default', result.output)
		self.assertNotIn('Test Type Boolean', result.output)
		self.assertNotIn('Test Type Integer', result.output)

		result = self.runner.invoke(cli, ['wp', 'option', 'get', 'sail_vars_test', '--format=json'])
		self.assertEqual(result.exit_code, 0)
		vars = json.loads(result.output)

		self.assertEqual(vars['test_simple'], '1')
		self.assertEqual(vars['test_default'], '1')
		self.assertEqual(vars['test_type_bool'], False)
		self.assertEqual(vars['test_type_int'], 789)

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_009_blueprint_define(self):
		result = self.runner.invoke(cli, ['blueprint', 'test_define.yaml'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Updating wp-config.php constants', result.output)

		# Test results
		result = self.runner.invoke(cli, ['wp', 'config', 'get', 'TEST_BOOLEAN', '--format=json'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('true', result.output)

		result = self.runner.invoke(cli, ['wp', 'config', 'get', 'TEST_INTEGER', '--format=json'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('123', result.output)

		result = self.runner.invoke(cli, ['wp', 'config', 'get', 'TEST_STRING', '--format=json'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('"This is a string"', result.output)

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_009_blueprint_plugins(self):
		result = self.runner.invoke(cli, ['blueprint', 'test_plugins.yaml'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Installing plugins', result.output)
		self.assertIn('wporg/debug-bar=latest', result.output)
		self.assertIn('wporg/classic-editor=latest', result.output)
		self.assertIn('thirdparty/hello-dolly', result.output)

		# Test results
		result = self.runner.invoke(cli, ['wp', 'plugin', 'list', '--skip-themes', '--skip-plugins', '--format=json'])
		self.assertEqual(result.exit_code, 0)

		plugins = json.loads(result.output)
		plugins = [(p['name'], p['status']) for p in plugins]

		self.assertIn(('hello-dolly', 'active'), plugins)
		self.assertIn(('classic-editor', 'active'), plugins)
		self.assertIn(('debug-bar', 'active'), plugins)

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_010_blueprint_themes(self):
		result = self.runner.invoke(cli, ['blueprint', 'test_themes.yaml'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Installing themes', result.output)
		self.assertIn('wporg/twentyten=latest', result.output)
		self.assertIn('wporg/twentyeleven=latest', result.output)
		self.assertIn('thirdparty/publish', result.output)

		# Test results
		result = self.runner.invoke(cli, ['wp', 'theme', 'list', '--skip-themes', '--skip-plugins', '--format=json'])
		self.assertEqual(result.exit_code, 0)

		themes = json.loads(result.output)
		themes = [(t['name'], t['status']) for t in themes]

		self.assertIn(('twentyten', 'inactive'), themes)
		self.assertIn(('twentyeleven', 'inactive'), themes)
		self.assertIn(('publish', 'active'), themes)

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_011_blueprint_options(self):
		# Set an option
		result = self.runner.invoke(cli, ['wp', 'option', 'update', 'test_delete_option', 'delete me'])
		self.assertEqual(result.exit_code, 0)

		result = self.runner.invoke(cli, ['blueprint', 'test_options.yaml'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Applying options', result.output)

		# Test results
		result = self.runner.invoke(cli, ['wp', 'option', 'get', 'blogname', '--format=json'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('"Blueprints Are Fun"', result.output)

		result = self.runner.invoke(cli, ['wp', 'option', 'get', 'blogdescription', '--format=json'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('"Don&#039;t forget to change your tagline!"', result.output)

		result = self.runner.invoke(cli, ['wp', 'option', 'get', 'some_custom_option', '--format=json'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('{"foo":"bar"}', result.output)

		# Never existed
		result = self.runner.invoke(cli, ['wp', 'option', 'get', 'test_delete_unknown_option'])
		self.assertEqual(result.exit_code, 1)

		# Should not exist
		result = self.runner.invoke(cli, ['wp', 'option', 'get', 'test_delete_option'])
		self.assertEqual(result.exit_code, 1)

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_012_blueprint_fail2ban(self):
		result = self.runner.invoke(cli, ['blueprint', 'test_fail2ban.yaml'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Installing fail2ban', result.output)
		self.assertIn('Configuring fail2ban rules', result.output)

		# Running again should not install.
		result = self.runner.invoke(cli, ['blueprint', 'test_fail2ban.yaml'])
		self.assertEqual(result.exit_code, 0)
		self.assertNotIn('Installing fail2ban', result.output)
		self.assertIn('Configuring fail2ban rules', result.output)

		# Make sure mu-plugin is installed
		result = self.runner.invoke(cli, ['wp', 'plugin', 'list', '--skip-themes', '--skip-plugins', '--format=json'])
		self.assertEqual(result.exit_code, 0)
		plugins = json.loads(result.output)
		plugins = [(p['name'], p['status']) for p in plugins]
		self.assertIn(('0-sail-auth-syslog', 'must-use'), plugins)

		# TODO: Check active jails and dpkg status when we have ssh [command]

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_013_blueprint_postfix(self):
		result = self.runner.invoke(cli, ['blueprint', 'postfix.yaml',
			'--smtp-host=smtp.gmail.com',
			'--smtp-port=578',
			'--smtp-username=foo',
			'--smtp-password=bar',
			'--from-name=From',
			'--from-email=me@example.org',
		])

		self.assertEqual(result.exit_code, 0)
		self.assertIn('Installing postfix', result.output)
		self.assertIn('Configuring postfix', result.output)

		# Running again should not install.
		result = self.runner.invoke(cli, ['blueprint', 'postfix.yaml',
			'--smtp-host=smtp.gmail.com',
			'--smtp-port=578',
			'--smtp-username=foo',
			'--smtp-password=bar',
			'--from-name=A Unit Test',
			'--from-email=me@example.org',
		])

		self.assertEqual(result.exit_code, 0)
		self.assertNotIn('Installing postfix', result.output)
		self.assertIn('Configuring postfix', result.output)

		# Make sure mu-plugin is installed
		result = self.runner.invoke(cli, ['wp', 'plugin', 'list', '--skip-themes', '--skip-plugins', '--format=json'])
		self.assertEqual(result.exit_code, 0)
		plugins = json.loads(result.output)
		plugins = [(p['name'], p['status']) for p in plugins]
		self.assertIn(('1-sail-mail-from', 'must-use'), plugins)

		result = self.runner.invoke(cli, ['download', '-y', 'wp-content'])
		self.assertEqual(result.exit_code, 0)

		with open('wp-content/mu-plugins/1-sail-mail-from.php', 'r') as f:
			contents = f.read()

		self.assertIn('A Unit Test', contents)
		self.assertIn('me@example.org', contents)

		# TODO: Check configs and dpkg status when we have ssh [command]

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_014_blueprint_dns(self):
		# Make sure domain doesn't exist.
		result = self.runner.invoke(cli, ['domain', 'delete', 'saildemo.com'])

		result = self.runner.invoke(cli, ['blueprint', 'test_dns.yaml', '--domain=saildemo.com'])
		self.assertEqual(result.exit_code, 1)
		self.assertIn('domain does not exist', result.output)

		result = self.runner.invoke(cli, ['domain', 'add', 'saildemo.com'])
		self.assertEqual(result.exit_code, 0)

		result = self.runner.invoke(cli, ['blueprint', 'test_dns.yaml', '--domain=saildemo.com'])
		self.assertEqual(result.exit_code, 0)
		self.assertNotIn('domain does not exist', result.output)

		self.assertIn('Creating A record for foo.saildemo.com', result.output)
		self.assertIn('Creating A record for bar.saildemo.com', result.output)
		self.assertIn('Creating MX record for baz.saildemo.com', result.output)
		self.assertIn('Creating CNAME record for qux.saildemo.com', result.output)
		self.assertIn('Blueprint applied successfully', result.output)

		# Applying again should not alter the records.
		result = self.runner.invoke(cli, ['blueprint', 'test_dns.yaml', '--domain=saildemo.com'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Skipping A record for foo.saildemo.com', result.output)
		self.assertIn('Skipping A record for bar.saildemo.com', result.output)
		self.assertIn('Skipping MX record for baz.saildemo.com', result.output)
		self.assertIn('Skipping CNAME record for qux.saildemo.com', result.output)
		self.assertIn('Blueprint applied successfully', result.output)

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	def test_015_profile_config(self):
		with open('.sail/config.json') as f:
			config = json.load(f)

		result = self.runner.invoke(cli, ['profile', 'key'])
		self.assertEqual(result.exit_code, 0)
		self.assertEqual(result.stdout, config['profile_key'])

		result = self.runner.invoke(cli, ['profile', 'key', '--header'])
		self.assertEqual(result.exit_code, 0)
		self.assertEqual(result.stdout, 'X-Sail-Profile: %s' % config['profile_key'])

	@unittest.skipIf(work_in_progress, 'Work in progress!')
	@patch('curses.wrapper')
	def test_016_profile(self, wrapper):
		with open('.sail/config.json') as f:
			config = json.load(f)

		result = self.runner.invoke(cli, ['profile', self.home])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Server:', result.output)
		self.assertIn('Host:', result.output)
		self.assertIn('Request:', result.output)
		self.assertIn('Downloading profile from', result.output)
		self.assertIn('Cleaning up production', result.output)
		self.assertIn('Profile saved to', result.output)

		wrapper.assert_called_once()
		totals = wrapper.call_args.kwargs['totals']

		self.assertGreater(totals['timestamp'], 1633423829)
		self.assertGreater(totals['queries'], 1)
		self.assertGreater(totals['ct'], 1)
		self.assertGreater(totals['wt'], 1)
		self.assertGreater(totals['mu'], 1)
		self.assertGreater(totals['pmu'], 1)
		self.assertEqual(totals['method'], 'GET')
		self.assertIn('SAIL_NO_CACHE=', totals['request_uri'])

		data = wrapper.call_args.kwargs['data']
		self.assertEqual(data['main()']['ct'], 1)
		self.assertGreater(data['main()']['wt'], 1)
		self.assertGreater(data['main()']['mu'], 1)
		self.assertGreater(data['main()']['pmu'], 1)

		self.assertGreater(data['main()']['excl_wt'], 1)
		self.assertGreater(data['main()']['excl_pmu'], 1)

		self.assertIn('wp', data['main()']['children'])
		self.assertIn('do_action#init', data['main()']['children'])

	def test_999_destroy(self):
		result = self.runner.invoke(cli, ['destroy', '-y'])
		self.assertEqual(result.exit_code, 0)

if __name__ == '__main__':
	unittest.main()



__title__ = 'sail'
__description__ = 'CLI tool to deploy and manage WordPress applications on DigitalOcean'
__version__ = '0.9.15'
__author__ = 'Konstantin Kovshenin'
__author_email__ = 'kovshenin@gmail.com'
__url__ = 'https://sailed.io'
__license__ = 'GPLv3'
__copyright__ = 'Copyright 2021 Konstantin Kovshenin'



from sail import cli

import unittest
from unittest.mock import Mock, patch
from click.testing import CliRunner

_find_root = Mock(return_value='/path/to/project')
_get_sail_config = Mock(return_value={
	"app_id": "foo",
	"secret": "bar",
	"url": "https://foobar.sailed.io/",
	"login_url": "https://foobar.sailed.io/wp-login.php",
	"version": "0.9.8"
})
_request = Mock()
_rmtree = Mock()

@patch('sail.util.find_root', _find_root)
@patch('sail.util.get_sail_config', _get_sail_config)
@patch('sail.util.request', _request)
@patch('shutil.rmtree', _rmtree)
class TestDestroy(unittest.TestCase):
	def setUp(self):
		_find_root.reset_mock()
		_get_sail_config.reset_mock()
		_request.reset_mock()
		_rmtree.reset_mock()

	def test_confirm(self):
		runner = CliRunner()
		result = runner.invoke(cli, ['destroy'])
		self.assertEqual(result.exit_code, 1)
		self.assertIn('Are you sure?', result.output)
		self.assertIn('Aborted!', result.output)
		_request.assert_not_called()
		_rmtree.assert_not_called()

	def test_destroy_input(self):
		runner = CliRunner()
		result = runner.invoke(cli, ['destroy'], input='y')
		self.assertEqual(result.exit_code, 0)
		_request.assert_called_once_with('/destroy/', method='DELETE')
		self.assertIn('destroyed successfully', result.output)
		_rmtree.assert_called_once_with('/path/to/project/.sail')

	def test_destroy_y(self):
		runner = CliRunner()
		result = runner.invoke(cli, ['destroy', '-y'])
		self.assertEqual(result.exit_code, 0)
		_request.assert_called_once_with('/destroy/', method='DELETE')
		self.assertIn('destroyed successfully', result.output)
		_rmtree.assert_called_once_with('/path/to/project/.sail')

	def test_help(self):
		runner = CliRunner()
		result = runner.invoke(cli, ['destroy', '--help'])
		self.assertEqual(result.exit_code, 0)
		self.assertIn('Usage:', result.output)
		_request.assert_not_called()
		_rmtree.assert_not_called()



from sail import cli, util, __version__

import json
import subprocess, os
import click
import shutil
from prettytable import PrettyTable

@cli.command()
@click.option('--provider-token', help='Your DigitalOcean API token, must be read-write. You can set a default token with: sail config provider-token <token>')
@click.option('--size', default='s-1vcpu-1gb-intel', help='The Droplet size, defaults to s-1vcpu-1gb-intel. To get a full list of available sizes and pricing run: sail sizes')
@click.option('--region', default='ams3', help='The region to deploy to, defaults to ams3. For a full list of available regions run: sail regions')
@click.option('--email', help='The admin e-mail address. You can set the default e-mail address with: sail config email <email>')
@click.option('--force', '-f', is_flag=True)
def init(provider_token, email, size, region, force):
	'''Initialize and provision a new project'''
	root = util.find_root()

	if root and os.path.exists(root + '/.sail'):
		raise click.ClickException('This ship has already sailed. Pick another one or remove the .sail directory.')

	files = os.listdir(path='.')
	if files and not force:
		raise click.ClickException('This project directory is not empty, can not init here. Override with --force, can destroy existing files.')

	if not provider_token:
		provider_token = util.get_sail_default('provider-token')

	if not provider_token:
		raise click.ClickException('You need to provide a DigitalOcean API token with --provider-token, or set a default one with: sail config provider-token <token>')

	if not email:
		email = util.get_sail_default('email')

	if not email:
		raise click.ClickException('You need to provide an admin e-mail address with --email, or set a default one with: sail config email <e-mail>')

	click.secho('# Initializing', bold=True)

	app = util.request('/init/', json={
		'provider_token': provider_token,
		'size': size,
		'email': email,
		'region': region,
	}, anon=True)

	app_id = app['app_id']
	private_key = app['private_key']
	public_key = app['public_key']

	click.echo('- Init successful, application id: %s' % app_id)
	os.mkdir('.sail')
	root = util.find_root()

	click.echo('- Writing SSH keys to .sail/ssh.key')
	with open('%s/.sail/ssh.key' % root, 'w+') as f:
		f.write(private_key)
	os.chmod('%s/.sail/ssh.key' % root, 0o600)

	with open('%s/.sail/ssh.key.pub' % root, 'w+') as f:
		f.write(public_key)
	os.chmod('%s/.sail/ssh.key.pub' % root, 0o644)

	click.echo('- Writing .sail/config.json')
	with open('%s/.sail/config.json' % root, 'w+') as f:
		json.dump({
			'app_id': app_id,
			'secret': app['secret'],
			'url': app['url'],
			'login_url': app['login_url'],
			'profile_key': app['profile_key'],
			'version': __version__,
		}, f, indent='\t')

	click.echo()
	click.secho('# Provisioning servers', bold=True)

	response = util.request('/provision/', method='POST')
	task_id = response['task_id']

	click.echo('- Provision scheduled successfully, waiting...')

	try:
		data = util.wait_for_task(task_id, 600, 5)
		if data['status'] != 'ready':
			raise Exception()
	except:
		raise click.ClickException('Provisioning failed. Please try again later.')

	click.echo('- Writing server keys to .sail/known_hosts')
	f = open('%s/.sail/known_hosts' % root, 'w+')
	r = subprocess.run(['ssh-keyscan', '-t', 'rsa,ecdsa', '-H', '%s.sailed.io' % app_id], stdout=f, stderr=subprocess.DEVNULL)
	f.close()

	click.echo()
	click.secho('# Downloading files from production', bold=True)

	args = ['-r']
	source = 'root@%s.sailed.io:/var/www/public/' % app_id
	destination = '%s/' % root
	returncode, stdout, stderr = util.rsync(args, source, destination)

	if returncode != 0:
		raise click.ClickException('An error occurred during download. Please try again.')

	click.echo('- Files download completed')

	# Create a local empty wp-contents/upload directory
	content_dir = '%s/wp-content/uploads' % root
	if not os.path.exists(content_dir):
		os.mkdir(content_dir)

	click.echo()
	click.secho('# Success. The ship has sailed!', bold=True)

	click.echo()
	click.echo('- URL: %s' % app['url'])
	click.echo('- Login: %s' % app['login_url'])
	click.echo('- Username: %s' % app['credentials']['username'])
	click.echo('- Password: %s (change me!)' % app['credentials']['password'])

	click.echo()
	click.echo('- SSH/SFTP Access Details')
	click.echo('- Host: %s.sailed.io' % app_id)
	click.echo('- Port: 22')
	click.echo('- Username: root')
	click.echo('- SSH Key: .sail/ssh.key')
	click.echo('- To open an interactive shell run: sail ssh')

	click.echo()
	click.echo('For support and documentation visit sailed.io')

@click.argument('path', nargs=-1, required=True)
@cli.command(context_settings=dict(ignore_unknown_options=True))
def blueprint(path):
	'''Run a blueprint file against your application'''
	import re
	import yaml
	import pathlib
	import json

	root = util.find_root()
	sail_config = util.get_sail_config()

	_args = path
	arguments = []
	options = {}
	flags = []

	for arg in _args:
		if not arg.startswith('-'):
			arguments.append(arg)
			continue

		# option or flag
		if '=' in arg:
			k, v = arg.split('=')
			options[k] = v
		else:
			flags.append(arg)

	path = arguments[0] # TODO: Support multiple paths
	path = pathlib.Path(path)

	# Try Sail's internal library of BPs
	if not path.exists() and path.parent == pathlib.Path('.'):
		path = pathlib.Path(__file__).parent / 'blueprints' / path.name

	if not path.exists():
		raise click.ClickException('File does not exist')

	if not path.name.endswith('.yml') and not path.name.endswith('.yaml'):
		raise click.ClickException('Blueprint files must be .yml or .yaml')

	with path.open() as f:
		s = f.read()

	def _parse_variables(match):
		name = match.group(1).strip()
		if name in vars:
			return json.dumps(vars[name])
		return None

	# Load user variables and fill from command line arguments if possible.
	y = yaml.safe_load(s)
	vars = {}
	for var in y.get('vars', []):
		option = var.get('option')
		_type = str

		_map = {
			'str': str,
			'string': str,
			'int': int,
			'integer': int,
			'float': float,
			'bool': bool,
			'boolean': bool,
		}

		if var.get('type') and var.get('type') in _map.keys():
			_type = _map[var.get('type')]

		if options.get(option):
			value = options.get(option)
		else:
			value = click.prompt(var['prompt'], default=var.get('default', None), type=_type)

		if _type == bool and type(value) is not bool:
			truthy = ['yes', 'y', 'true', '1', 'affirmative']
			value = True if value.lower() in truthy else False

		elif _type == int or _type == float:
			try:
				value = _type(value)
			except ValueError:
				raise click.ClickException('Could not convert %s to %s' % (repr(value), repr(_type)))

		vars[var['name']] = value

	# Reload with substitutions.
	s = re.sub(r'\${{([^}]+?)}}', _parse_variables, s)
	y = yaml.safe_load(s)

	click.echo('# Applying blueprint: %s' % path.name)
	response = util.request('/blueprint/', method='POST', json={'blueprint': y})
	task_id = response['task_id']

	util.wait_for_task(task_id, timeout=600, interval=5)

	click.echo('- Blueprint applied successfully')

@cli.command()
@click.option('--yes', '-y', is_flag=True, help='Force Y on overwriting local copy')
def destroy(yes):
	'''Shutdown and destroy the production droplet'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	if not yes:
		click.confirm('All droplet data will be scrubbed and irretrievable. Are you sure?', abort=True)

	app_id = sail_config['app_id']

	click.echo()
	click.echo('# Destroying application')

	data = util.request('/destroy/', method='DELETE')

	click.echo('- Droplet destroyed successfully')
	click.echo('- Removing .sail/*')

	shutil.rmtree(root + '/.sail')

@cli.command()
def sizes():
	'''Get available droplet sizes'''
	click.echo()
	click.echo('# Getting available droplet sizes')

	data = util.request('/sizes/', anon=True)
	t = PrettyTable(['Size', 'Price', 'Description'])

	for slug, size in data.items():
		t.add_row([slug, size['price_monthly'], size['description']])

	t.align = 'l'
	t.sortby = 'Price'
	click.echo(t.get_string())

@cli.command()
def regions():
	'''Get available deployment regions'''
	click.echo()
	click.echo('# Getting available regions')

	data = util.request('/regions/', anon=True)
	t = PrettyTable(['Slug', 'Name'])

	for slug, region in data.items():
		t.add_row([slug, region['name']])

	t.align = 'l'
	click.echo(t.get_string())



from sail import cli, util

import requests, json, os, subprocess, time
import click, hashlib, pathlib, shutil
from datetime import datetime

@cli.command()
@click.argument('path', nargs=1, required=True)
@click.option('--yes', '-y', is_flag=True, help='Skip the AYS message and force yes')
@click.option('--skip-db', is_flag=True, help='Do not import the database')
@click.option('--skip-uploads', is_flag=True, help='Do not import uploads')
def restore(path, yes, skip_db, skip_uploads):
	'''Restore your application files, uploads and database from a backup file'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	path = pathlib.Path(path)
	if not path.exists():
		raise click.ClickException('File does not exist')

	if path.name.endswith('.sql.gz') or path.name.endswith('.sql'):
		raise click.ClickException('Looks like a database-only backup. Try: sail db import')

	if not path.name.endswith('.tar.gz'):
		raise click.ClickException('Doesn\'t look like a backup file')

	if not yes:
		click.confirm('This will restore a full backup to production. Continue?', abort=True)

	click.echo('# Restoring backup')

	app_id = sail_config['app_id']
	backups_dir = pathlib.Path(root + '/.backups')
	backups_dir.mkdir(parents=True, exist_ok=True)
	progress_dir = pathlib.Path(backups_dir / ('.%s.progress' % hashlib.sha256(os.urandom(32)).hexdigest()[:8]))
	progress_dir.mkdir()

	database_filename = 'database.%s.sql.gz' % hashlib.sha256(os.urandom(32)).hexdigest()[:8]

	click.echo('- Extracting backup files')

	p = subprocess.Popen([
		'tar', ('-xzvf' if util.debug() else '-xzf'), path.resolve(), '--directory', progress_dir.resolve()
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred during backup. Please try again.')

	for x in progress_dir.iterdir():
		if x.name not in ['www', 'database.sql.gz', 'uploads']:
			shutil.rmtree(progress_dir)
			raise click.ClickException('Unexpected file in backup archive: %s' % x.name)

	if skip_uploads:
		click.echo('- Skipping uploads')
	else:
		click.echo('- Importing uploads')

		args = ['-rtl', '--delete', '--rsync-path', 'sudo -u www-data rsync']
		source = '%s/uploads/' % progress_dir
		destination = 'root@%s.sailed.io:/var/www/uploads/' % app_id
		returncode, stdout, stderr = util.rsync(args, source, destination, default_filters=False)

		if returncode != 0:
			shutil.rmtree(progress_dir)
			raise click.ClickException('An error occurred during restore. Please try again.')

	click.echo('- Importing application files')

	args = ['-rtl', '--delete', '--rsync-path', 'sudo -u www-data rsync']
	source = '%s/www/' % progress_dir
	destination = 'root@%s.sailed.io:/var/www/public/' % app_id
	returncode, stdout, stderr = util.rsync(args, source, destination)

	if returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred during restore. Please try again.')

	if skip_db:
		click.echo('- Skipping database import')
	else:
		click.echo('- Uploading database backup')

		args = ['-t']
		source = '%s/database.sql.gz' % progress_dir
		destination = 'root@%s.sailed.io:/var/www/%s' % (app_id, database_filename)
		returncode, stdout, stderr = util.rsync(args, source, destination, default_filters=False)

		if returncode != 0:
			shutil.rmtree(progress_dir)
			raise click.ClickException('An error occurred in rsync. Please try again.')

		click.echo('- Importing database into MySQL')

		# TODO: Maybe do an atomic import which deletes tables that no longer exist
		# by doing a rename.
		p = subprocess.Popen(['ssh',
			'-i', '%s/.sail/ssh.key' % root,
			'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
			'-o', 'IdentitiesOnly=yes',
			'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
			'root@%s.sailed.io' % sail_config['app_id'],
			'docker exec sail bash -c "zcat /var/www/%s | mysql -uroot wordpress"' % database_filename,
		])

		while p.poll() is None:
			util.loader()

		if p.returncode != 0:
			shutil.rmtree(progress_dir)
			raise click.ClickException('An error occurred in SSH. Please try again.')

		click.echo('- Cleaning up production')

		p = subprocess.Popen(['ssh',
			'-i', '%s/.sail/ssh.key' % root,
			'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
			'-o', 'IdentitiesOnly=yes',
			'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
			'root@%s.sailed.io' % sail_config['app_id'],
			'docker exec sail rm /var/www/%s' % database_filename
		])

		while p.poll() is None:
			util.loader()

		if p.returncode != 0:
			shutil.rmtree(progress_dir)
			raise click.ClickException('An error occurred in SSH. Please try again.')

	shutil.rmtree(progress_dir)
	click.echo('- Backup restored successfully. Your local copy may be out of date.')

@cli.command()
def backup():
	'''Backup your production files and database to your local .backups directory'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	click.echo('# Backing up')

	app_id = sail_config['app_id']
	backups_dir = pathlib.Path(root + '/.backups')
	backups_dir.mkdir(parents=True, exist_ok=True)
	progress_dir = pathlib.Path(backups_dir / ('.%s.progress' % hashlib.sha256(os.urandom(32)).hexdigest()[:8]))
	(progress_dir / 'www').mkdir(parents=True)
	(progress_dir / 'uploads').mkdir()

	database_filename = 'database.%s.sql.gz' % hashlib.sha256(os.urandom(32)).hexdigest()[:8]

	click.echo('- Downloading application files')

	args = ['-rtl', '--copy-dest', '%s/' % root]
	source = 'root@%s.sailed.io:/var/www/public/' % app_id
	destination = '%s/www/' % progress_dir
	returncode, stdout, stderr = util.rsync(args, source, destination)

	if returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred during backup. Please try again.')

	click.echo('- Downloading uploads')

	args = ['-rtl', '--copy-dest', '%s/wp-content/uploads/' % root]
	source = 'root@%s.sailed.io:/var/www/uploads/' % app_id
	destination = '%s/uploads/' % progress_dir
	returncode, stdout, stderr = util.rsync(args, source, destination, default_filters=False)

	if returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred during backup. Please try again.')

	click.echo('- Exporting WordPress database')

	p = subprocess.Popen(['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'docker exec sail bash -c "mysqldump --quick --single-transaction --default-character-set=utf8mb4 -uroot wordpress | gzip -c9 > /var/www/%s"' % database_filename
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred in SSH. Please try again.')

	click.echo('- Export completed, downloading database')

	args = ['-t']
	source = 'root@%s.sailed.io:/var/www/%s' % (sail_config['app_id'], database_filename)
	destination = '%s/database.sql.gz' % progress_dir
	returncode, stdout, stderr = util.rsync(args, source, destination, default_filters=False)

	if returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred in rsync. Please try again.')

	click.echo('- Cleaning up production')

	p = subprocess.Popen(['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'docker exec sail rm /var/www/%s' % database_filename
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred in SSH. Please try again.')

	timestamp = datetime.now().strftime('%Y-%m-%d-%H%M%S.tar.gz')
	target = pathlib.Path(backups_dir / timestamp)

	click.echo('- Archiving and compressing backup files')

	p = subprocess.Popen([
		'tar', ('-cvzf' if util.debug() else '-czf'), target.resolve(), '-C', progress_dir.resolve(), '.'
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred during backup. Please try again.')

	shutil.rmtree(progress_dir)

	click.echo('- Backup completed at .backups/%s' % timestamp)



from sail import cli, util

import subprocess, time
import click, pathlib

def _get_extend_filters(paths, prefix=None):
	root = util.find_root()

	extend_filters = []
	if not paths:
		return []

	if prefix:
		prefixed_root = root.rstrip('/') + '/' + prefix.strip('/')

	for entry in paths:
		_entry = pathlib.Path(entry)

		# Requested root directory, remove all filters.
		if str(_entry.resolve()) == root:
			return []

		try:
			relative = _entry.resolve().relative_to(root)
		except ValueError:
			raise click.ClickException('Could not resolve path: %s' % entry)

		if prefix:
			try:
				relative = _entry.resolve().relative_to(prefixed_root)
			except ValueError:
				continue # The item is in root, but not in relative root, skip

			if str(relative) == '.':
				continue

		for parent in reversed(relative.parents):
			if str(parent) == '.':
				continue
			extend_filters.append('+ /%s' % parent)

		if _entry.is_dir():
			extend_filters.append('+ /%s/***' % relative)
		else:
			extend_filters.append('+ /%s' % relative)

	# Skip everything else.
	if len(extend_filters) > 0:
		extend_filters.append('- *')

	return extend_filters

@cli.command()
@click.argument('path', nargs=-1, required=False)
@click.option('--with-uploads', is_flag=True, help='Include the wp-content/uploads directory')
@click.option('--dry-run', is_flag=True, help='Show changes about to be deployed to production')
def deploy(with_uploads, dry_run, path):
	'''Deploy your working copy to production. If path is not specified then all application files are deployed.'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	app_id = sail_config['app_id']
	release_name = str(int(time.time()))

	if dry_run:
		click.echo('# Comparing files')

		destination = 'root@%s.sailed.io:/var/www/public/' % app_id
		source = '%s/' % root
		files = _diff(source, destination, _get_extend_filters(path))
		empty = True

		colors = {'created': 'green', 'deleted': 'red', 'updated': 'yellow'}
		labels = {'created': 'New', 'deleted': 'Delete', 'updated': 'Update'}
		for op in ['created', 'updated', 'deleted']:
			for filename in files[op]:
				empty = False
				click.secho('- %s: %s' % (labels[op], filename), fg=colors[op])

		# TODO: Compare uploads if requested --with-uploads

		if empty:
			click.echo('- No changes')

		return

	click.echo('# Deploying to production')
	click.echo('- Preparing release directory')

	p = subprocess.Popen(['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'mkdir -p /var/www/releases/%s && rsync -rogtl /var/www/public/ /var/www/releases/%s' % (release_name, release_name)
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		raise click.ClickException('An error occurred in SSH. Please try again.')

	click.echo('- Uploading application files to production')

	returncode, stdout, stderr = util.rsync(
		args=['-rtl', '--rsync-path', 'sudo -u www-data rsync',
			'--copy-dest', '/var/www/public/', '--delete'],
		source='%s/' % root,
		destination='root@%s.sailed.io:/var/www/releases/%s' % (app_id, release_name),
		extend_filters=_get_extend_filters(path)
	)

	if returncode != 0:
		raise click.ClickException('An error occurred during upload. Please try again.')

	if with_uploads:
		click.echo('- Uploading wp-content/uploads')

		# Send uploads to production
		returncode, stdout, stderr = util.rsync(
			args=['-rtl', '--rsync-path', 'sudo -u www-data rsync', '--delete'],
			source='%s/wp-content/uploads/' % root,
			destination='root@%s.sailed.io:/var/www/uploads/' % app_id,
			default_filters=False,
			extend_filters=_get_extend_filters(path, 'wp-content/uploads')
		)

		if returncode != 0:
			raise click.ClickException('An error occurred during upload. Please try again.')

	click.echo('- Requesting Sail API to deploy: %s' % release_name)

	data = util.request('/deploy/', json={'release': release_name})
	task_id = data.get('task_id')

	if not task_id:
		raise click.ClickException('Could not obain a deploy task_id.')

	click.echo('- Scheduled successfully, waiting for deploy')

	try:
		data = util.wait_for_task(task_id, timeout=300, interval=5)
	except:
		raise click.ClickException('Deploy failed')

	click.echo('- Successfully deployed %s' % release_name)

@cli.command()
@click.argument('release', required=False, type=int, nargs=1)
@click.option('--releases', is_flag=True, help='Get a list of valid releases to rollback to')
def rollback(release=None, releases=False):
	'''Rollback production to a previous release'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	if releases or not release:
		data = util.request('/rollback')
		click.echo('# Available releases:')
		if data.get('releases', []):
			for r in data.get('releases', []):
				flags = '(current)' if r == data.get('current_release') else ''
				click.echo('- %s %s' % (r, flags))

			click.echo()
			click.echo('Rollback with: sail rollback <release>')
		else:
			click.echo('- No releases found, perhaps you should deploy something')

		return

	if release:
		release = str(release)

	click.echo()
	click.echo('# Rolling back')

	if release:
		click.echo('- Requesting Sail API to rollback: %s' % release)
	else:
		click.echo('- Requesting Sail API to rollback')

	data = util.request('/rollback/', json={'release': release})
	task_id = data.get('task_id')
	rollback_release = data.get('release')

	if not task_id:
		raise click.ClickException('Could not obain a rollback task_id.')

	click.echo('- Scheduled successfully, waiting for rollback')

	try:
		data = util.wait_for_task(task_id, timeout=300, interval=5)
	except:
		raise click.ClickException('Rollback failed')

	click.echo('- Successfully rolled back to %s' % rollback_release)

@cli.command()
@click.argument('path', nargs=-1, required=False)
@click.option('--yes', '-y', is_flag=True, help='Force Y on overwriting local copy')
@click.option('--with-uploads', is_flag=True, help='Include the wp-content/uploads directory')
@click.option('--delete', is_flag=True, help='Delete files from local copy that do not exist on production')
@click.option('--dry-run', is_flag=True, help='Show changes about to be downloaded to the working copy')
def download(path, yes, with_uploads, delete, dry_run):
	'''Download files from production to your working copy'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	if not yes and not dry_run:
		click.confirm('Downloading files from production may overwrite '
			+ 'your local copy. Continue?',
			abort=True
		)

	app_id = sail_config['app_id']
	delete = ['--delete'] if delete else []

	if dry_run:
		click.echo('# Comparing files')

		source = 'root@%s.sailed.io:/var/www/public/' % app_id
		destination = '%s/' % root
		files = _diff(source, destination, _get_extend_filters(path))
		empty = True

		colors = {'created': 'green', 'deleted': 'red', 'updated': 'yellow'}
		labels = {'created': 'New', 'deleted': 'Delete', 'updated': 'Update'}
		for op in ['created', 'updated', 'deleted']:
			for filename in files[op]:
				empty = False
				click.secho('- %s: %s' % (labels[op], filename), fg=colors[op])

		# TODO: Compare uploads if requested --with-uploads

		if empty:
			click.echo('- No changes')

		return

	click.echo('# Downloading application files from production')

	returncode, stdout, stderr = util.rsync(
		args=['-rtl'] + delete,
		source='root@%s.sailed.io:/var/www/public/' % app_id,
		destination='%s/' % root,
		extend_filters=_get_extend_filters(path)
	)

	if returncode != 0:
		raise click.ClickException('An error occurred during download. Please try again.')

	if with_uploads:
		click.echo('- Downloading wp-content/uploads')

		# Download uploads from production
		returncode, stdout, stderr = util.rsync(
			args=['-rtl'] + delete,
			source='root@%s.sailed.io:/var/www/uploads/' % app_id,
			destination='%s/wp-content/uploads/' % root,
			default_filters=False,
			extend_filters=_get_extend_filters(path, 'wp-content/uploads')
		)

		if returncode != 0:
			raise click.ClickException('An error occurred during download. Please try again.')

	click.echo('- Files download completed')

def _diff(source, destination, extend_filters=[]):
	'''Compare application files between environments'''
	root = util.find_root()

	if source == destination:
		raise click.ClickException('Can not compare apples to apples')

	args = ['-rlci', '--delete', '--dry-run']
	returncode, stdout, stderr = util.rsync(args, source, destination, extend_filters=extend_filters)

	if returncode != 0:
		raise click.ClickException('An error occurred in rsync. Please try again.')

	files = {
		'created': [],
		'deleted': [],
		'updated': [],
	}

	if util.debug():
		util.dlog(stdout)

	for line in stdout.splitlines():
		change = line[:11]
		path = line[12:]

		if len(change) < 11:
			continue

		# See --itemize-changes https://linux.die.net/man/1/rsync
		# Deleted
		if change.strip() == '*deleting':
			files['deleted'].append(path)
			continue

		# Newly created file.
		if change == '<f+++++++++' or change == '>f+++++++++':
			files['created'].append(path)
			continue

		# Newly created directory.
		if change == 'cd+++++++++':
			files['created'].append(path)
			continue

		# Checksum different.
		if change[:3] == '<fc' or change[:3] == '>fc':
			files['updated'].append(path)
			continue

		# Permission change
		if change[5] == 'p':
			files['updated'].append(path)
			continue

	return files



from sail import cli, util

import os, subprocess
import click
import hashlib
import pathlib
from datetime import datetime

@cli.group()
def db():
	'''Import and export MySQL databases, or spawn an interactive shell'''
	pass

@db.command()
def cli():
	'''Open an interactive MySQL shell on the production host'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	click.echo('Spawning an interactive MySQL shell at %s.sailed.io' % sail_config['app_id'])

	os.execlp('ssh', 'ssh', '-t',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'docker exec -it sail sudo -u www-data wp --path=/var/www/public db cli'
	)

@db.command(name='import')
@click.argument('path', nargs=1, required=True)
def import_cmd(path):
	'''Import a local .sql or .sql.gz file to the production MySQL database'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	path = pathlib.Path(path).resolve()
	if not path.exists():
		raise click.ClickException('File does not exist')

	if not path.name.endswith('.sql') and not path.name.endswith('.sql.gz'):
		raise click.ClickException('This does not look like a .sql or .sql.gz file')

	temp_name = '%s.%s' % (hashlib.sha256(os.urandom(32)).hexdigest()[:8], path.name)
	is_gz = path.name.endswith('.sql.gz')

	click.echo('# Importing WordPress database')
	click.echo('- Uploading database file to production')

	args = ['-t']
	source = path
	destination = 'root@%s.sailed.io:/var/www/%s' % (sail_config['app_id'], temp_name)
	returncode, stdout, stderr = util.rsync(args, source, destination, default_filters=False)

	if returncode != 0:
		raise click.ClickException('An error occurred in rsync. Please try again.')

	# TODO: Maybe do an atomic import which deletes tables that no longer exist
	# by doing a rename.

	click.echo('- Importing database into MySQL')
	cat_bin = 'zcat' if is_gz else 'cat'

	p = subprocess.Popen(['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'docker exec sail bash -c "%s /var/www/%s | mysql -uroot wordpress"' % (cat_bin, temp_name)
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		raise click.ClickException('An error occurred in SSH. Please try again.')

	click.echo('- Cleaning up production')

	p = subprocess.Popen(['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'docker exec sail rm /var/www/%s' % temp_name
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		raise click.ClickException('An error occurred in SSH. Please try again.')

	click.echo('- Database imported')

@db.command()
def export():
	'''Export the production database to a local .sql.gz file'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	backups_dir = pathlib.Path(root + '/.backups')
	backups_dir.mkdir(parents=True, exist_ok=True)
	filename = datetime.now().strftime('%Y-%m-%d-%H%M%S.sql.gz')

	click.echo('# Exporting WordPress database')

	p = subprocess.Popen(['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'docker exec sail bash -c "mysqldump --quick --single-transaction --default-character-set=utf8mb4 -uroot wordpress | gzip -c9 > /var/www/%s"' % filename
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		raise click.ClickException('An error occurred in SSH. Please try again.')

	click.echo('- Export completed, downloading')

	args = ['-t']
	source = 'root@%s.sailed.io:/var/www/%s' % (sail_config['app_id'], filename)
	destination = '%s/%s' % (backups_dir, filename)
	returncode, stdout, stderr = util.rsync(args, source, destination, default_filters=False)

	if returncode != 0:
		raise click.ClickException('An error occurred in rsync. Please try again.')

	click.echo('- Cleaning up production')

	p = subprocess.Popen(['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'docker exec sail rm /var/www/%s' % filename
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		raise click.ClickException('An error occurred in SSH. Please try again.')

	click.echo('- Database export saved to .backups/%s' % filename)



from sail import cli, util

import requests, json, os, subprocess, time
import click

@cli.group()
def domain():
	'''Add, remove and update domains associated with your site'''
	pass

@domain.command()
def list():
	'''List domains associated with your site'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	response = util.request('/domains/')

	click.secho('# Domains', bold=True)
	for domain, data in response.items():
		flags = []
		for flag in ['internal', 'https', 'primary']:
			if data.get(flag, None):
				flags.append(flag)

		sflags = ''
		if flags:
			sflags = ' (%s)' % ', '.join(flags)

		click.echo('- ' + domain + sflags)

@domain.command()
@click.argument('domain', nargs=1)
@click.option('--force', is_flag=True, help='Make primary again, even if the domain is already set as the primary one')
@click.option('--skip-replace', is_flag=True, help='Skip running search-replace routines for home, siteurl, and other URLs')
def make_primary(domain, force, skip_replace):
	'''Set a domain as primary, update siteurl/home, search-replace all links'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	click.echo('# Updating primary domain')

	response = util.request('/domains/make-primary/', json={
		'domain': domain,
		'force': bool(force),
		'skip_replace': bool(skip_replace),
	})
	task_id = response['task_id']

	click.echo('- Scheduled make-primary for %s' % domain)
	click.echo('- Waiting for make-primary to complete')

	util.wait_for_task(task_id, timeout=300, interval=5)

	click.echo('- Primary domain updated successfully')

@domain.command()
@click.argument('domains', nargs=-1)
@click.option('--agree-tos', is_flag=True)
def make_https(domains, agree_tos):
	'''Request and install SSL certificates for domains'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	if not agree_tos:
		click.echo('Let\'s Encrypt ToS: https://community.letsencrypt.org/tos')
		click.confirm('Do you agree to the Let\'s Encrypt ToS?', abort=True)

	if not domains:
		raise click.ClickException('At least one domain is required')

	click.echo('# Requesting and installing SSL for domains')

	response = util.request('/domains/make-https/', json={'domains': domains})
	task_id = response['task_id']

	click.echo('- Scheduled make-https for %s' % ', '.join(domains))
	click.echo('- Waiting for make-https to complete')

	util.wait_for_task(task_id, timeout=300, interval=5)

	click.echo('- SSL certificates installed')
	click.echo('- Don\'t forget to: sail domain make-primary')

@domain.command()
@click.option('--skip-dns', is_flag=True, help='Do not add or update DNS records')
@click.argument('domains', nargs=-1)
def add(domains, skip_dns):
	'''Add a new domain, with DNS records pointing to your site'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	if not domains:
		raise click.ClickException('At least one domain is required')

	response = util.request('/domains/', json={'domains': domains, 'skip_dns': skip_dns})

	for domain, data in response['feedback'].items():
		click.echo()
		click.echo('# %s' % domain)
		for line in data:
			click.echo('- %s' % line)

@domain.command()
@click.option('--skip-dns', is_flag=True, help='Do not delete DNS records')
@click.argument('domains', nargs=-1)
def delete(domains, skip_dns):
	'''Delete a domain and all DNS records'''
	'''Add a new domain, with DNS records pointing to your site'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	response = util.request('/domains/', json={'domains': domains, 'skip_dns': skip_dns}, method='DELETE')

	for domain, data in response['feedback'].items():
		click.echo()
		click.echo('# %s' % domain)
		for line in data:
			click.echo('- %s' % line)



from setuptools import setup
from setuptools import find_packages
import os, pathlib

import sail

p = pathlib.Path(__file__)
p = p.parent.resolve()
requirements = str(p) + '/requirements.txt'

with open(requirements) as f:
	install_reqs = f.read().splitlines()

setup(
	name='sailed.io',
	version=sail.__version__,
	description=sail.__description__,
	author=sail.__author__,
	author_email=sail.__author_email__,
	url=sail.__url__,
	install_requires=install_reqs,
	include_package_data=True,
	packages=find_packages(),
	entry_points={
		'console_scripts': ['sail=sail:cli']
	},
)



from .__version__ import __version__, __title__, __description__, __url__
from .__version__ import __author__, __author_email__, __license__, __copyright__

API_BASE = 'https://sailed.io/api/1.0'

from sail import util

import click

@click.group()
@click.version_option(__version__, '--version', '-v', message='%(version)s')
@click.option('--debug', '-d', is_flag=True, help='Enable verbose debug logging')
def cli(debug):
	util.debug(debug)

# Command groups
from sail import provision, deploy, database, backups, domains, misc, profiling



from sail import cli, __version__

import unittest
from click.testing import CliRunner

class TestVersion(unittest.TestCase):
	def test_version(self):
		runner = CliRunner()
		result = runner.invoke(cli, ['--version'])
		self.assertEqual(result.exit_code, 0)
		self.assertEqual(result.output, __version__ + '\n')

	def test_version_short(self):
		runner = CliRunner()
		result = runner.invoke(cli, ['-v'])
		self.assertEqual(result.exit_code, 0)
		self.assertEqual(result.output, __version__ + '\n')
